import findspark
findspark.init("/usr/local/Cellar/spark-2.4.6-bin-hadoop2.7")

# import os
# os.environ["HADOOP_USER_NAME"] = "huangning"
# os.environ["PYSPARK_SUBMIT_ARGS"] = '--jars /Users/bytedance/ByteCode/' \
#                                     'magellan_megaspark/megaspark/' \
#                                     'libs/xgboost4j-0.72.jar,/Users/' \
#                                     'bytedance/ByteCode/magellan_megaspark/' \
#                                     'megaspark/libs/' \
#                                     'xgboost4j-spark-0.72.jar pyspark-shell'

from pyspark.sql import DataFrame
from pyspark.sql import SparkSession
from megaspark.sql.megaframe import MegaFrame
from pyspark.sql.types import StructType, StructField
from pyspark.sql.types import ByteType, ShortType, \
    IntegerType, LongType, FloatType, DoubleType, BooleanType, StringType


spark = SparkSession\
    .builder\
    .appName("mega sql")\
    .master("local")\
    .getOrCreate()
# spark.sparkContext.addPyFile("/Users/bytedance"
#                              "/ByteCode/magellan_megaspark"
#                              "/megaspark/libs/sparkxgb.zip")
# from sparkxgb import XGBoostEstimator


print("Welcome Mega Spark ～")


def show_func():

    print("+-------------------+")
    print("|sql method         |")
    print("+-------------------+")
    print("|1. read_csv        |")
    print("|2. sql             |")
    print("+-------------------+")
    print("|MegaFrame method   |")
    print("+-------------------+")
    print("|1. head            |")
    print("|2. table_alias     |")
    print("|3. sort_values     |")
    print("|4. fillna          |")
    print("+-------------------+")


def read_csv(input_path, header=True, feats_info={}):
    """Read a comma-separated values (csv) file into spark DataFrame.

    Parameters
    ----------
    input_path : str
        Any valid string path is acceptable.
        The string should be the path of HDFS.

    header : bool, default=True
        If True, the real column names of the
        data frame will be shown. if False, the
        column name of the DataFrame will be
        displayed as _c0, _c1, _c2 and so on.

    feat_info : dict, default={}
        Mapping of features and feature types

    Returns
    --------
    spark_df : DataFrame
        The result of the read csv file.


    Examples
    ----------
    >>> input_path = "/path/to/sample.csv"
    >>> data_df =  mg.read_csv(input_path, True)
    >>> data_df.mega.head(3)
      PassengerId Survived Pclass  ...     Fare Cabin Embarked
    0           1        0      3  ...     7.25  None        S
    1           2        1      1  ...  71.2833   C85        C
    2           3        1      3  ...    7.925  None        S
    3           4        1      1  ...     53.1  C123        S
    4           5        0      3  ...     8.05  None        S
    """
    if len(feats_info) == 0:
        return spark.read.option("header", header).csv(input_path)

    schema = []
    for feat_name, feat_type in feats_info.items():

        if feat_type == "bool":
            schema.append(StructField(feat_name, DoubleType()))
        elif feat_type == "string":
            schema.append(StructField(feat_name, StringType()))
        elif feat_type == "byte":
            schema.append(StructField(feat_name, ByteType()))
        elif feat_type == "short":
            schema.append(StructField(feat_name, ShortType()))
        elif feat_type == "int":
            schema.append(StructField(feat_name, IntegerType()))
        elif feat_type == "long":
            schema.append(StructField(feat_name, LongType()))
        elif feat_type == "float":
            schema.append(StructField(feat_name, FloatType()))
        elif feat_type == "double":
            schema.append(StructField(feat_name, DoubleType()))
        else:
            raise Exception("Current type can not handle")

    schema_info = StructType(schema)
    return spark.read.option("header", header).schema(schema_info).csv(input_path)


def sql(sql_query):
    """Read a comma-separated values (csv) file into spark DataFrame.

    Parameters
    ----------
    sqlquery : str
        Query statement


    Returns
    --------
    spark_df : DataFrame
        The result of the given query.


    Examples
    ----------
    >>> df.mega.table_alias("people")
    >>> data_df =  mg.sql("SELECT field1 AS f1, field2 AS f2 from people")
    >>> data_df.mega.head(5)
          PassengerId Survived Pclass  ...     Fare Cabin Embarked
    0           1        0      3  ...     7.25  None        S
    1           2        1      1  ...  71.2833   C85        C
    2           3        1      3  ...    7.925  None        S
    3           4        1      1  ...     53.1  C123        S
    4           5        0      3  ...     8.05  None        S

    Note
    ------
    Alias the table by `df.table_alias("tabel_name")` before using SQL queries
    """
    return spark.sql(sql_query)


def train_test_split(df, test_size, random_state):
    """The sample set is divided into training set and Validation set

    Parameters
    ----------
    df : spark DataFrame
        Sample set to be splited.

    test_size : float
        The value should be between 0.0 and 1.0 and represent the proportion
        of the dataset to include in the test split.

    random_state : int
        Controls the shuffling applied to the data before applying the split.


    Returns
    --------
    train_df : DataFrame
        Training sample set

    test_df : DataFrame
        Testing sample set


    Examples
    ----------
    >>> df.mega.head(5)
      PassengerId Survived Pclass  ...     Fare Cabin Embarked
    0           1        0      3  ...     7.25  None        S
    1           2        1      1  ...  71.2833   C85        C
    2           3        1      3  ...    7.925  None        S
    3           4        1      1  ...     53.1  C123        S
    4           5        0      3  ...     8.05  None        S
    >>> trainDF, testDF = tm.train_test_split(df, test_size=0.3, random_state=99)
    >>> trainDF.mega.head(5)
          PassengerId Survived Pclass  ...     Fare Cabin Embarked
    0           1        0      3  ...     7.25  None        S
    1          10        1      2  ...  30.0708  None        C
    2         100        0      2  ...       26  None        S
    3         103        0      1  ...  77.2875   D26        S
    4         104        0      3  ...   8.6542  None        S
    """
    train_size = 1 - test_size
    train_df, test_df = df.randomSplit(
        [train_size, test_size], seed=random_state)
    return train_df, test_df


@property
def ext_magellan(self):
    """Converts Spark DataFrame into MegaFrame
    """
    return MegaFrame(self)


DataFrame.mega = ext_magellan



from pyspark.sql import SparkSession
from pyspark.ml import Pipeline
from pyspark.sql.functions import col
from pyspark.ml.feature import StringIndexer, VectorAssembler
from pyspark.sql.types import StructType, StructField, DoubleType, StringType
from pyspark.ml.feature import StringIndexer, VectorAssembler
from megaspark.ml.xgboost import XGBoostEstimator
# from megaspark.tomega import XGBoostEstimator

class XGBoostClassifier:

    def __init__(self, feat_name, label_name, pred_name):

        self.label_name = label_name
        self.pred_name = pred_name
        self.feat_name = feat_name

        self.xgb = XGBoostEstimator(
            featuresCol=feat_name,
            labelCol=label_name,
            predictionCol=pred_name
        )

    def fit(self, df):

        # In order to convert the nominal values into numeric
        # ones we need to define aTransformer for each column:
        sexIndexer = StringIndexer() \
            .setInputCol("Sex") \
            .setOutputCol("SexIndex") \
            .setHandleInvalid("keep")

        cabinIndexer = StringIndexer() \
            .setInputCol("Cabin") \
            .setOutputCol("CabinIndex") \
            .setHandleInvalid("keep")

        embarkedIndexer = StringIndexer() \
            .setInputCol("Embarked") \
            .setOutputCol("EmbarkedIndex") \
            .setHandleInvalid("keep")

        vectorAssembler = VectorAssembler() \
            .setInputCols(["Pclass", "SexIndex", "Age",
                           "SibSp", "Parch", "Fare",
                           "CabinIndex", "EmbarkedIndex"]) \
            .setOutputCol("features")

        pipeline = Pipeline().setStages([sexIndexer,
                                         cabinIndexer, embarkedIndexer,
                                         vectorAssembler, self.xgb])

        # get a new transformer
        self.model = pipeline.fit(df)

    def predict_proba(self, df):

        return self.model.transform(df).select(self.label_name, "probabilities", self.pred_name)




#
# Copyright (c) 2018 by Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# from pyspark import SparkContext, keyword_only
# from pyspark.ml.common import _java2py

from pyspark.ml.param.shared import HasFeaturesCol, HasLabelCol, HasPredictionCol, HasWeightCol, HasCheckpointInterval
from pyspark.ml.util import JavaMLWritable
from pyspark.ml.wrapper import JavaEstimator, JavaModel
from megaspark.ml.util import XGBoostReadable


# <------------------增加JavaPredictionModel----------------->

# from pyspark import since
from functools import wraps


def since(version):
    """
    A decorator that annotates a function to append the version of Spark the function was added.
    """
    import re
    indent_p = re.compile(r'\n( +)')

    def deco(f):
        indents = indent_p.findall(f.__doc__)
        indent = ' ' * (min(len(m) for m in indents) if indents else 0)
        f.__doc__ = f.__doc__.rstrip() + "\n\n%s.. versionadded:: %s" % (indent, version)
        return f
    return deco


class JavaPredictionModel:
    """
    (Private) Java Model for prediction tasks (regression and classification).
    To be mixed in with class:`pyspark.ml.JavaModel`
    """

    @property
    @since("2.1.0")
    def numFeatures(self):
        """
        Returns the number of features the model was trained on. If unknown, returns -1
        """
        return self._call_java("numFeatures")


#<------------  增加keyword ----------------------->
def keyword_only(func):
    """
    A decorator that forces keyword arguments in the wrapped method
    and saves actual input keyword arguments in `_input_kwargs`.

    .. note:: Should only be used to wrap a method where first arg is `self`
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if len(args) > 0:
            raise TypeError("Method %s forces keyword arguments." % func.__name__)
        self._input_kwargs = kwargs
        return func(self, **kwargs)
    return wrapper

#<------------  增加_java2py  --------------------->
import sys
if sys.version >= '3':
    long = int
    unicode = str

from py4j.protocol import Py4JJavaError
from py4j.java_gateway import JavaObject
from py4j.java_collections import JavaArray, JavaList

from pyspark import RDD, SparkContext
from pyspark.serializers import PickleSerializer
from pyspark.sql import DataFrame, SQLContext


_picklable_classes = [
    'LinkedList',
    'SparseVector',
    'DenseVector',
    'DenseMatrix',
    'Rating',
    'LabeledPoint',
]


def _java2py(sc, r, encoding="bytes"):
    if isinstance(r, JavaObject):
        clsName = r.getClass().getSimpleName()
        # convert RDD into JavaRDD
        if clsName != 'JavaRDD' and clsName.endswith("RDD"):
            r = r.toJavaRDD()
            clsName = 'JavaRDD'

        if clsName == 'JavaRDD':
            jrdd = sc._jvm.SerDe.javaToPython(r)
            return RDD(jrdd, sc)

        if clsName == 'DataFrame':
            return DataFrame(r, SQLContext(sc))

        if clsName in _picklable_classes:
            r = sc._jvm.SerDe.dumps(r)
        elif isinstance(r, (JavaArray, JavaList)):
            try:
                r = sc._jvm.SerDe.dumps(r)
            except Py4JJavaError:
                pass  # not pickable

    if isinstance(r, (bytearray, bytes)):

        r = PickleSerializer().loads(bytes(r), encoding=encoding)

    return r

#<----------------  --------------------->


class JavaParamsOverrides(object):
    """
    Mixin for overriding methods derived from JavaParams.
    """

    # Define a fix similar to SPARK-10931 (For Spark <2.3)
    def _create_params_from_java(self):
        """
        Create params that are defined in the Java obj but not here
        """
        java_params = list(self._java_obj.params())
        from pyspark.ml.param import Param
        for java_param in java_params:
            java_param_name = java_param.name()
            if not hasattr(self, java_param_name):
                param = Param(self, java_param_name, java_param.doc())
                setattr(param, "created_from_java_param", True)
                setattr(self, java_param_name, param)
                self._params = None  # need to reset so self.params will discover new params

    # Backport SPARK-10931 (For Spark <2.3)
    def _transfer_params_from_java(self):
        """
        Transforms the embedded params from the companion Java object.
        """
        sc = SparkContext._active_spark_context
        for param in self.params:
            if self._java_obj.hasParam(param.name):
                java_param = self._java_obj.getParam(param.name)
                # SPARK-14931: Only check set params back to avoid default params mismatch.
                if self._java_obj.isSet(java_param):
                    value = _java2py(sc, self._java_obj.getOrDefault(java_param))
                    self._set(**{param.name: value})
                # SPARK-10931: Temporary fix for params that have a default in Java
                if self._java_obj.hasDefault(java_param) and not self.isDefined(param):
                    value = _java2py(sc, self._java_obj.getDefault(java_param)).get()
                    self._setDefault(**{param.name: value})

    # Override the "_from_java" method, so we can read our objects.
    @classmethod
    def _from_java(cls, java_stage):
        """
        Given a Java object, create and return a Python wrapper of it.
        """

        # Create a new instance of this stage.
        py_stage = cls()

        # Load information from java_stage to the instance.
        py_stage._java_obj = java_stage
        py_stage._create_params_from_java()
        py_stage._resetUid(java_stage.uid())
        py_stage._transfer_params_from_java()

        return py_stage


class XGBoostEstimator(JavaParamsOverrides, JavaEstimator, HasCheckpointInterval, HasFeaturesCol, HasLabelCol,
                       HasPredictionCol, HasWeightCol, JavaMLWritable, XGBoostReadable):
    """
    A PySpark implementation of ml.dmlc.xgboost4j.scala.spark.XGBoostEstimator.
    """

    @keyword_only
    def __init__(self,
                 # General Params
                 checkpoint_path="", checkpointInterval=-1, missing=None, nthread=1, nworkers=1, silent=0,
                 use_external_memory=False,

                 # Column Params
                 baseMarginCol="baseMargin", featuresCol="features", labelCol="label", predictionCol="prediction",
                 weightCol="weight",

                 # Booster Params
                 base_score=0.5, booster="gbtree", eval_metric="error", num_class=2, num_round=2,
                 objective="binary:logistic", seed=None,

                 # Tree Booster Params
                 alpha=0.0, colsample_bytree=1.0, colsample_bylevel=1.0, eta=0.3, gamma=0.0, grow_policy='depthwise',
                 max_bin=256, max_delta_step=0.0, max_depth=6, min_child_weight=1.0, reg_lambda=0.0,
                 scale_pos_weight=1.0, sketch_eps=0.03, subsample=1.0, tree_method="auto",

                 # Dart Booster Params
                 normalize_type="tree", rate_drop=0.0, sample_type="uniform", skip_drop=0.0,

                 # Linear Booster Params
                 lambda_bias=0.0):

        super(XGBoostEstimator, self).__init__()
        self._java_obj = self._new_java_obj("ml.dmlc.xgboost4j.scala.spark.XGBoostEstimator", self.uid)
        self._create_params_from_java()
        self._setDefault(
            # Column Params
            featuresCol="features", labelCol="label", predictionCol="prediction",
            weightCol="weight", baseMarginCol="baseMargin",

            # Booster Params
            objective="binary:logistic", eval_metric="error", num_round=2)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @keyword_only
    def setParams(self,
                  # General Params
                  checkpoint_path="", checkpointInterval=-1, missing=None, nthread=1, nworkers=1, silent=0,
                  use_external_memory=False,

                  # Column Params
                  baseMarginCol="baseMargin", featuresCol="features", labelCol="label", predictionCol="prediction",
                  weightCol="weight",

                  # Booster Params
                  base_score=0.5, booster="gbtree", eval_metric="error", num_class=2, num_round=2,
                  objective="binary:logistic", seed=None,

                  # Tree Booster Params
                  alpha=0.0, colsample_bytree=1.0, colsample_bylevel=1.0, eta=0.3, gamma=0.0, grow_policy='depthwise',
                  max_bin=256, max_delta_step=0.0, max_depth=6, min_child_weight=1.0, reg_lambda=0.0,
                  scale_pos_weight=1.0, sketch_eps=0.03, subsample=1.0, tree_method="auto",

                  # Dart Booster Params
                  normalize_type="tree", rate_drop=0.0, sample_type="uniform", skip_drop=0.0,

                  # Linear Booster Params
                  lambda_bias=0.0):

        kwargs = self._input_kwargs_processed()
        return self._set(**kwargs)

    def _input_kwargs_processed(self):
        """
        Until consensus on parameter names can be achieved, we must rename kwargs which would break python.
        """

        kwargs = self._input_kwargs
        if "reg_lambda" in kwargs:
            kwargs["lambda"] = kwargs.pop("reg_lambda")

        return kwargs

    def _create_model(self, java_model):
        """
        Create the correct python object for the model type.
        """

        java_package = java_model.getClass().getName()
        java_class = java_package.split('.')[-1]

        if java_class == 'XGBoostClassificationModel':
            return XGBoostClassificationModel(java_model)
        elif java_class == 'XGBoostRegressionModel':
            return XGBoostRegressionModel(java_model)
        else:
            raise NotImplementedError("This XGBoost model type cannot loaded into Python currently: %r"
                                      % java_class)


class XGBoostClassificationModel(JavaParamsOverrides, JavaModel, JavaPredictionModel, JavaMLWritable, XGBoostReadable):
    """
    A PySpark implementation of ml.dmlc.xgboost4j.scala.spark.XGBoostClassificationModel.
    """

    def __init__(self, java_model=None):
        """
        Override the __init__ from JavaModel.
        """

        super(XGBoostClassificationModel, self).__init__(java_model)
        if java_model is not None:
            # Get parameters only present in the model object.
            self._create_params_from_java()

            self._resetUid(java_model.uid())

            # Transfer parameter values from java object.
            self._transfer_params_from_java()

    @property
    def numClasses(self):
        """
        Number of classes (values which the label can take).
        """
        return self._call_java("numClasses")

    def setThresholds(self, value):
        """
        Sets the value of :py:attr:`thresholds`.
        """
        return self._set(thresholds=value)

    def getThresholds(self):
        """
        Gets the value of thresholds or its default value.
        """
        return self.getOrDefault(self.thresholds)

    def setRawPredictionCol(self, value):
        """
        Sets the value of :py:attr:`rawPredictionCol`.
        """
        return self._set(rawPredictionCol=value)

    def getRawPredictionCol(self):
        """
        Gets the value of rawPredictionCol or its default value.
        """
        return self.getOrDefault(self.rawPredictionCol)


class XGBoostRegressionModel(JavaParamsOverrides, JavaModel, JavaPredictionModel, JavaMLWritable, XGBoostReadable):
    """
    A PySpark implementation of ml.dmlc.xgboost4j.scala.spark.XGBoostRegressionModel.
    """

    def __init__(self, java_model=None):
        """
        Override the __init__ from JavaModel.
        """

        super(XGBoostRegressionModel, self).__init__(java_model)
        if java_model is not None:
            # Get parameters only present in the model object.
            self._create_params_from_java()

            self._resetUid(java_model.uid())

            # Transfer parameter values from java object.
            self._transfer_params_from_java()



#
# Copyright (c) 2018 by Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pyspark.ml.util import JavaMLReadable, JavaMLReader


class XGBoostReadable(JavaMLReadable):
    """
    Mixin for instances that provide XGBoostReader.
    """

    @classmethod
    def read(cls):
        """Returns an XGBoostReader instance for this class."""
        return XGBoostReader(cls)


class XGBoostReader(JavaMLReader):
    """
    A reader for XGBoost objects.
    """

    def __init__(self, clazz):
        self._clazz = clazz
        self._jread = self._load_java_obj(clazz).read()

    @classmethod
    def _java_loader_class(cls, clazz):
        """
        Returns the full class name of the Java XGBoost class. In this case,
        we prepend "ml.dmlc.xgboost4j.scala.spark" to the name of the class.
        """

        # Deal with Python locations for XGBoost objects.
        if clazz.__name__ in ("XGBoostEstimator", "XGBoostClassificationModel", "XGBoostRegressionModel"):
            java_package = clazz.__module__.replace("sparkxgb", "ml.dmlc.xgboost4j.scala.spark")

        # Our Pipeline object calls down to Spark's Pipeline object.
        elif clazz.__name__ == "XGBoostPipeline":
            return "org.apache.spark.ml.Pipeline"

        # Our PipelineModel object calls down to Spark's PipelineModel object.
        elif clazz.__name__ == "XGBoostPipelineModel":
            return "org.apache.spark.ml.PipelineModel"

        # Allow default PySpark objects to be loaded.
        else:
            java_package = clazz.__module__.replace("pyspark", "org.apache.spark")

        if clazz.__name__ in ("Pipeline", "PipelineModel", "XGBoostEstimator", "XGBoostClassificationModel",
                              "XGBoostRegressionModel"):
            # Remove the last dangling package name, e.g. "pipeline" or "xgboost".
            java_package = ".".join(java_package.split(".")[0:-1])

        return java_package + "." + clazz.__name__



import megaspark.tomega as tm
from megaspark.ml.mega_xgboost import XGBoostClassifier

if __name__ == "__main__":
    input_path = "/Users/bytedance/ByteCode/magellan_megaspark/data/train.csv"
    data_df = tm.read_csv(input_path, header=True, feats_info={"PassengerId": "double",
                                                               "Survived": "double",
                                                               "Pclass": "double",
                                                               "Name": "string",
                                                               "Sex": "string",
                                                               "Age": "double",
                                                               "SibSp": "double",
                                                               "Parch": "double",
                                                               "Ticket": "string",
                                                               "Fare": "double",
                                                               "Cabin": "string",
                                                               "Embarked": "string"})

    data_df = data_df.mega.fillna({"PassengerId": 0,
                                   "Survived": 0,
                                   "Pclass": 0,
                                   "Name": "0",
                                   "Sex": "0",
                                   "Age": 0,
                                   "SibSp": 0,
                                   "Parch": 0,
                                   "Ticket": "0",
                                   "Fare": 0,
                                   "Cabin": "0",
                                   "Embarked": "0"})

    # data_df = data_df.na.fill(0)
    data_df.mega.table_alias("student")

    # # sql method
    # sample_df = tm.sql("select * from student limit 10")

    # split dataset into trainDF and testDF
    trainDF, testDF = tm.train_test_split(data_df, test_size=0.3, random_state=99)

    # model training
    xgb_clf = XGBoostClassifier("features", "Survived", "prediction")
    xgb_clf.fit(trainDF)
    res = xgb_clf.predict_proba(trainDF)
    print(res.mega.head(5))



from pyspark.sql.types import ByteType, ShortType, \
    IntegerType, LongType, FloatType, DoubleType, BooleanType, StringType


class MegaFrame(object):

    def __init__(self, df):
        self._df = df
        self.feat_info = [(field.name, field.dataType)
                          for field in self._df.schema.fields]

    def __getitem__(self, *args):
        """Read the DataFrame according to the column names.

        Examples
        ----------
        >>> data_df.mega[["PassengerId", "Survived"]].mega.head(3)
          PassengerId Survived
        0           1        0
        1           2        1
        2           3        1
        """
        col_names = args[0]
        return self._df.select(col_names)

    def head(self, n: int = 5):

        """This function returns the first `n` rows for
        the DataFrame. For negative values of `n`,
        this function returns all rows except
        the last `n` rows, equivalent to ``df[:-n]``.

        Parameters
        ----------
        n : int, default=5
            Number of rows to select.

        Returns
        ---------
        data_df : pandas DataFrame
            The first `n` rows of the caller object.

        Examples
        ----------
        >>> data_df.mega.head(5)
              PassengerId Survived Pclass  ...     Fare Cabin Embarked
        0           1        0      3  ...     7.25  None        S
        1           2        1      1  ...  71.2833   C85        C
        2           3        1      3  ...    7.925  None        S
        3           4        1      1  ...     53.1  C123        S
        4           5        0      3  ...     8.05  None        S
        :param n:
        """
        return self._df.toPandas().head(n)

    def sort_values(self, *col_names, ascending=True):
        """Sort by one or more column names of the data frame

        Parameters
        ----------
        col_names: tuple
            tuple of column names to sort by.

        ascending: bool, default=True
            If True, sort values in ascending order, otherwise descending.

        Returns
        ---------
        data_df: spark DataFrame
            Sorted DataFrame

        Examples
        ----------
        >>> data_df.mega.sort_values("PassengerId", "Survived",
        ... ascending=False).mega.head(3)
          PassengerId Survived Pclass  ...     Fare    Cabin Embarked
        0          99        1      2  ...       23     None        S
        1          98        1      1  ...  63.3583  D10 D12        C
        2          97        0      1  ...  34.6542       A5        C
        """
        return self._df.orderBy(*col_names, ascending=ascending)

    def fillna(self, cols_dict={}):
        """Fill in missing values with fill_value

        Parameters
        ------------
        cols_dict: dict, default={}
            key is colname, value is the missing value to fill in

        Returns
        ------------
        data_df: spark DataFrame
            DataFrame with no missing value

        Examples
        ----------
        >>> data_df.mega.head(5)
          PassengerId Survived Pclass  ...     Fare Cabin Embarked
        0           1        0      3  ...     7.25  None        S
        1           2        1      1  ...  71.2833   C85        C
        2           3        1      3  ...    7.925  None        S
        3           4        1      1  ...     53.1  C123        S
        4           5        0      3  ...     8.05  None        S
        >>> df = data_df.mega.fillna({"Survived": 0, "Cabin": "unknown"})
        >>> df.mega.head(5)
          PassengerId Survived Pclass  ...     Fare    Cabin Embarked
        0           1        0      3  ...     7.25  unknown        S
        1           2        1      1  ...  71.2833      C85        C
        2           3        1      3  ...    7.925  unknown        S
        3           4        1      1  ...     53.1     C123        S
        4           5        0      3  ...     8.05  unknown        S
        """

        if len(cols_dict) == 0:
            for name, dataType in self.feat_info:
                if isinstance(dataType, (
                        ByteType, ShortType, IntegerType, LongType)):
                    cols_dict[name] = 0
                elif isinstance(dataType, (FloatType, DoubleType)):
                    cols_dict[name] = 0.0
                elif isinstance(dataType, (StringType, BooleanType)):
                    cols_dict[name] = "0"
                else:
                    raise Exception("The missing value "
                                    "of the current data "
                                    "type cannot be handled")
        return self._df.na.fill(cols_dict)

    def table_alias(self, name):
        """Creates or replaces a local temporary view with DataFrame.
        The lifetime of this temporary table is tied to `SparkSession`

        Examples
        ----------
        >>> df.table_alias("people")
        """
        self._df.createOrReplaceTempView(name)



# coding: utf-8

from setuptools import find_packages, setup

import io
import os

# 包元信息
NAME = 'mega-spark'
DESCRIPTION = 'Some computing tools about feature tool and ml model.'
AUTHOR = 'huangning'

# 项目运行需要的依赖
REQUIRES = ["pyspark>=2.4.0,<2.4.7"
            ]

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except IOError:
    long_description = DESCRIPTION

setup(
      name=''+NAME,
      version='1.1.1',
      description=DESCRIPTION,
      long_description=long_description,
      long_description_content_type="text/markdown",
      author=AUTHOR,
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
      ],
      packages=find_packages(),
      install_requires=REQUIRES,
      python_requires='>=2.7,<=3.7.*',
      include_package_data=True
)



#
# Copyright (c) 2018 by Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pyspark.ml import Pipeline, PipelineModel
from pyspark.ml.base import Estimator, Transformer
from pyspark.ml.wrapper import JavaParams
from megaspark.ml.util import XGBoostReader


class JavaParamsXGBoost(JavaParams):
    """
    A class derived from pyspark.ml.wrapper.JavaParams which is able to load XGBoost stages.
    """

    @staticmethod
    def _from_java(java_stage):
        """
        Given a Java object, create and return a Python wrapper of it.
        Used for ML persistence.
        """

        def __get_class(clazz):
            """
            Loads Python class from its name.
            """
            parts = clazz.split('.')
            module = ".".join(parts[:-1])
            m = __import__(module)
            for comp in parts[1:]:
                m = getattr(m, comp)
            return m

        stage_name = java_stage.getClass().getName()
        class_name = stage_name.split('.')[-1]

        if class_name in ("XGBoostEstimator", "XGBoostClassificationModel", "XGBoostRegressionModel"):
            stage_name = stage_name.replace("ml.dmlc.xgboost4j.scala.spark", "sparkxgb")

        # Our Pipeline's Java object is just the default Spark one. (Cast all Pipelines to our version)
        elif class_name == "Pipeline":
            stage_name = "sparkxgb.XGBoostPipeline"

        # Our PipelineModel's Java object is just the default Spark one. (Cast all PipelineModels to our version)
        elif class_name == "PipelineModel":
            stage_name = "sparkxgb.XGBoostPipelineModel"

        else:
            stage_name = stage_name.replace("org.apache.spark", "pyspark")

        # Generate a default new instance from the stage_name class.
        py_type = __get_class(stage_name)

        if hasattr(py_type, "_from_java"):
            py_stage = py_type._from_java(java_stage)

        else:
            raise NotImplementedError("This Java stage cannot be loaded into Python currently: %r"
                                      % stage_name)
        return py_stage


class XGBoostPipeline(Pipeline):
    """
    A class derived from Pipeline which is able to read/write XGBoost estimator stages.
    """

    @classmethod
    def read(cls):
        """Returns an XGBoostReader instance for this class."""
        return XGBoostReader(cls)

    @classmethod
    def _from_java(cls, java_stage):
        """
        Given a Java Pipeline, create and return a Python wrapper of it.
        Used for ML persistence.
        """
        # Create a new instance of this stage.
        py_stage = cls()
        # Load information from java_stage to the instance.
        py_stages = [JavaParamsXGBoost._from_java(s) for s in java_stage.getStages()]
        py_stage.setStages(py_stages)
        py_stage._resetUid(java_stage.uid())
        return py_stage

    def _fit(self, dataset):
        """
        This method mirrors the default Pipeline _fit, the only difference is we return an XGBoostPipelineModel object.
        """
        stages = self.getStages()
        for stage in stages:
            if not (isinstance(stage, Estimator) or isinstance(stage, Transformer)):
                raise TypeError(
                    "Cannot recognize a pipeline stage of type %s." % type(stage))
        indexOfLastEstimator = -1
        for i, stage in enumerate(stages):
            if isinstance(stage, Estimator):
                indexOfLastEstimator = i
        transformers = []
        for i, stage in enumerate(stages):
            if i <= indexOfLastEstimator:
                if isinstance(stage, Transformer):
                    transformers.append(stage)
                    dataset = stage.transform(dataset)
                else:  # must be an Estimator
                    model = stage.fit(dataset)
                    transformers.append(model)
                    if i < indexOfLastEstimator:
                        dataset = model.transform(dataset)
            else:
                transformers.append(stage)
        return XGBoostPipelineModel(transformers)


class XGBoostPipelineModel(PipelineModel):
    """
    A class derived from PipelineModel which is able to read/write XGBoost transformer stages.
    """

    @classmethod
    def read(cls):
        """Returns an XGBoostReader instance for this class."""
        return XGBoostReader(cls)

    @classmethod
    def _from_java(cls, java_stage):
        """
        Given a Java PipelineModel, create and return a Python wrapper of it.
        Used for ML persistence.
        """
        # Load information from java_stage to the instance.
        py_stages = [JavaParamsXGBoost._from_java(s) for s in java_stage.stages()]
        # Create a new instance of this stage.
        py_stage = cls(py_stages)
        py_stage._resetUid(java_stage.uid())
        return py_stage
